#!/bin/bash
#
# cvgfilter     Filter for rejecting email messages with dangerous
#               attachments (sendmail extension).
#
# chkconfig: - 79 31
# description: Filter for rejecting email messages with dangerous attachments
#              via Milter library (sendmail extension). Sendmail needs to define
#              filter in sendmail.cf so read the doc please.
# processname: cvgfilter
# config:
# pidfile: /var/run/cvgfilter.pid

# Source function library.
. /etc/rc.d/init.d/functions

# Source networking configuration.
. /etc/sysconfig/network

# Check that networking is up.
[ ${NETWORKING} = "no" ] && exit 0

[ -f /usr/sbin/cvgfilter ] || exit 0

# Source cvgfilter configuration
if [ -f /etc/sysconfig/cvgfilter ] ; then
        . /etc/sysconfig/cvgfilter
fi

[ -z "$CVSOCKET" ] && CVSOCKET="/var/run/cvgfilter.sock"

RETVAL=0
prog="cvgfilter"

start() {
	# remove socket only if no cvgfilter is running
        if [ -e $CVSOCKET ]; then
		if [ -z "$(pidof $prog)" ]; then
			rm -f $CVSOCKET
			echo -n $"Stale socket removed: $CVSOCKET"
			success $"Stale socket removed: $CVSOCKET"
			echo
		else
			echo -n $"Starting $prog: already running "
			failure $"Starting $prog: already running "
			echo
			return 1
		fi
	fi
	echo -n $"Starting $prog: "
	umask 077
	daemon /usr/sbin/cvgfilter -p unix:$CVSOCKET
	RETVAL=$?
	echo
	[ $RETVAL -eq 0 ] && touch /var/lock/subsys/$prog
	return $RETVAL
}

stop() {
	# Stop daemons.
	echo -n $"Shutting down $prog: "
	killproc $prog
	RETVAL=$?
	echo
	[ $RETVAL -eq 0 ] && rm -f /var/lock/subsys/$prog
	return $RETVAL
}

# See how we were called.
case "$1" in
  start)
	start
	;;
  stop)
	stop
	;;
  restart|reload)
	stop
	start
	RETVAL=$?
	;;
  condrestart)
	if [ -f /var/lock/subsys/$prog ]; then
	    stop
	    start
	    RETVAL=$?
	fi
	;;
  status)
	status $prog
	RETVAL=$?
	;;
  *)
	echo $"Usage: $0 {start|stop|restart|condrestart|status}"
	exit 1
esac

exit $RETVAL
